# Copyright (c) 2010-2024 Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2008-2016, Neotion
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

# pylint: disable=broad-except
# pylint: disable=invalid-name
# pylint: disable=missing-function-docstring
# pylint: disable=missing-module-docstring
# pylint: disable=no-member
# pylint: disable=super-with-arguments

from sys import stderr
from time import time
from ..misc import hexdump


__all__ = ['SerialLogger']


class SerialLogger:
    """Serial port wrapper to log input/output data to a log file.
    """

    def __init__(self, *args, **kwargs):
        logpath = kwargs.pop('logfile', None)
        if not logpath:
            raise ValueError('Missing logfile')
        try:
            # pylint: disable=consider-using-with
            self._logger = open(logpath, "wt")
        except IOError as exc:
            print(f'Cannot log data to {logpath}: {exc}', file=stderr)
        self._last = time()
        self._log_init(*args, **kwargs)
        super(SerialLogger, self).__init__(*args, **kwargs)

    def open(self,):
        self._log_open()
        super(SerialLogger, self).open()

    def close(self):
        self._log_close()
        self._logger.close()
        super(SerialLogger, self).close()

    def read(self, size=1):
        data = super(SerialLogger, self).read(size)
        self._log_read(data)
        return data

    def write(self, data):
        if data:
            self._log_write(data)
        super(SerialLogger, self).write(data)

    def flush(self):
        self._log_flush()
        super(SerialLogger, self).flush()

    def reset_input_buffer(self):
        self._log_reset('I')
        super(SerialLogger, self).reset_input_buffer()

    def reset_output_buffer(self):
        self._log_reset('O')
        super(SerialLogger, self).reset_output_buffer()

    def send_break(self, duration=0.25):
        self._log_signal('BREAK', f'for {duration:.3f}')
        super(SerialLogger, self).send_break()

    def _update_break_state(self):
        self._log_signal('BREAK', self._break_state)
        super(SerialLogger, self)._update_break_state()

    def _update_rts_state(self):
        self._log_signal('RTS', self._rts_state)
        super(SerialLogger, self)._update_rts_state()

    def _update_dtr_state(self):
        self._log_signal('DTR', self._dtr_state)
        super(SerialLogger, self)._update_dtr_state()

    @property
    def cts(self):
        level = super(SerialLogger, self).cts
        self._log_signal('CTS', level)
        return level

    @property
    def dsr(self):
        level = super(SerialLogger, self).dsr
        self._log_signal('DSR', level)
        return level

    @property
    def ri(self):
        level = super(SerialLogger, self).ri
        self._log_signal('RI', level)
        return level

    @property
    def cd(self):
        level = super(SerialLogger, self).cd
        self._log_signal('CD', level)
        return level

    def in_waiting(self):
        count = super(SerialLogger, self).in_waiting()
        self._log_waiting(count)
        return count

    def _print(self, header, string=None):
        if self._logger:
            now = time()
            delta = (now-self._last)*1000
            self._last = now
            print(f'{header} ({delta:3.3f} ms):\n{string or ""}',
                  file=self._logger)
            self._logger.flush()

    def _log_init(self, *args, **kwargs):
        try:
            sargs = ', '.join(args)
            skwargs = ', '.join({f'{it[0]}={it[1]}' for it in kwargs.items()})
            self._print('NEW', f'  args: {sargs} {skwargs}')
        except Exception as exc:
            print(f'Cannot log init ({exc})', file=stderr)

    def _log_open(self):
        try:
            self._print('OPEN')
        except Exception as exc:
            print(f'Cannot log open ({exc})', file=stderr)

    def _log_close(self):
        try:
            self._print('CLOSE')
        except Exception as exc:
            print(f'Cannot log close ({exc})', file=stderr)

    def _log_read(self, data):
        try:
            self._print('READ', hexdump(data))
        except Exception as exc:
            print(f'Cannot log input data ({exc})', file=stderr)

    def _log_write(self, data):
        try:
            self._print('WRITE', hexdump(data))
        except Exception as exc:
            print(f'Cannot log output data ({exc})', data, file=stderr)

    def _log_flush(self):
        try:
            self._print('FLUSH')
        except Exception as exc:
            print(f'Cannot log flush action ({exc})', file=stderr)

    def _log_reset(self, type_):
        try:
            self._print('RESET BUFFER', type_)
        except Exception as exc:
            print(f'Cannot log reset buffer ({exc})', file=stderr)

    def _log_waiting(self, count):
        try:
            self._print('INWAITING', f'{count}')
        except Exception as exc:
            print(f'Cannot log inwaiting ({exc})', file=stderr)

    def _log_signal(self, name, value):
        try:
            self._print(name.upper(), str(value))
        except Exception as exc:
            print(f'Cannot log {name} ({exc})', file=stderr)



#!/usr/bin/env python3

"""Simple FTDI EEPROM configurator.
"""

# Copyright (c) 2019-2024, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from argparse import ArgumentParser, FileType
from io import StringIO
from logging import Formatter, StreamHandler, DEBUG, ERROR
from sys import exit as sys_exit, modules, stderr, stdout
from textwrap import fill
from traceback import format_exc
from pyftdi import FtdiLogger
from pyftdi.eeprom import FtdiEeprom
from pyftdi.ftdi import Ftdi
from pyftdi.misc import add_custom_devices, hexdump

# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements


def main():
    """Main routine"""
    debug = False
    try:
        argparser = ArgumentParser(description=modules[__name__].__doc__)
        argparser.add_argument('device', nargs='?', default='ftdi:///?',
                               help='serial port device name')

        files = argparser.add_argument_group(title='Files')
        files.add_argument('-i', '--input', type=FileType('rt'),
                           help='input ini file to load EEPROM content')
        files.add_argument('-l', '--load', default='all',
                           choices=('all', 'raw', 'values'),
                           help='section(s) to load from input file')
        files.add_argument('-o', '--output',
                           help='output ini file to save EEPROM content')
        files.add_argument('-V', '--virtual', type=FileType('r'),
                           help='use a virtual device, specified as YaML')

        device = argparser.add_argument_group(title='Device')
        device.add_argument('-P', '--vidpid', action='append',
                            help='specify a custom VID:PID device ID '
                                 '(search for FTDI devices)')
        device.add_argument('-M', '--eeprom',
                            help='force an EEPROM model')
        device.add_argument('-S', '--size', type=int,
                            choices=FtdiEeprom.eeprom_sizes,
                            help='force an EEPROM size')

        fmt = argparser.add_argument_group(title='Format')
        fmt.add_argument('-x', '--hexdump', action='store_true',
                         help='dump EEPROM content as ASCII')
        fmt.add_argument('-X', '--hexblock', type=int,
                         help='dump EEPROM as indented hexa blocks')

        config = argparser.add_argument_group(title='Configuration')
        config.add_argument('-s', '--serial-number',
                            help='set serial number')
        config.add_argument('-m', '--manufacturer',
                            help='set manufacturer name')
        config.add_argument('-p', '--product',
                            help='set product name')
        config.add_argument('-c', '--config', action='append',
                            help='change/configure a property as key=value '
                                 'pair')
        config.add_argument('--vid', type=lambda x: int(x, 16),
                            help='shortcut to configure the USB vendor ID')
        config.add_argument('--pid', type=lambda x: int(x, 16),
                            help='shortcut to configure the USB product ID')

        action = argparser.add_argument_group(title='Action')
        action.add_argument('-e', '--erase', action='store_true',
                            help='erase the whole EEPROM content')
        action.add_argument('-E', '--full-erase', action='store_true',
                            default=False,
                            help='erase the whole EEPROM content, including '
                                 'the CRC')
        action.add_argument('-u', '--update', action='store_true',
                            help='perform actual update, use w/ care')

        extra = argparser.add_argument_group(title='Extras')
        extra.add_argument('-v', '--verbose', action='count', default=0,
                           help='increase verbosity')
        extra.add_argument('-d', '--debug', action='store_true',
                           help='enable debug mode')
        args = argparser.parse_args()
        debug = args.debug

        if not args.device:
            argparser.error('Serial device not specified')

        loglevel = max(DEBUG, ERROR - (10 * args.verbose))
        loglevel = min(ERROR, loglevel)
        if debug:
            formatter = Formatter('%(asctime)s.%(msecs)03d %(name)-20s '
                                  '%(message)s', '%H:%M:%S')
        else:
            formatter = Formatter('%(message)s')
        FtdiLogger.set_formatter(formatter)
        FtdiLogger.set_level(loglevel)
        FtdiLogger.log.addHandler(StreamHandler(stderr))

        if args.virtual:
            # pylint: disable=import-outside-toplevel
            from pyftdi.usbtools import UsbTools
            # Force PyUSB to use PyFtdi test framework for USB backends
            UsbTools.BACKENDS = ('pyftdi.tests.backend.usbvirt', )
            # Ensure the virtual backend can be found and is loaded
            backend = UsbTools.find_backend()
            loader = backend.create_loader()()
            loader.load(args.virtual)

        try:
            add_custom_devices(Ftdi, args.vidpid, force_hex=True)
        except ValueError as exc:
            argparser.error(str(exc))

        eeprom = FtdiEeprom()
        eeprom.open(args.device, size=args.size, model=args.eeprom)
        if args.erase or args.full_erase:
            eeprom.erase()
        if args.input:
            eeprom.load_config(args.input, args.load)
        if args.serial_number:
            eeprom.set_serial_number(args.serial_number)
        if args.manufacturer:
            eeprom.set_manufacturer_name(args.manufacturer)
        if args.product:
            eeprom.set_product_name(args.product)
        for conf in args.config or []:
            if conf in ('?', 'help'):
                helpstr = ', '.join(sorted(eeprom.properties))
                print(fill(helpstr, initial_indent='  ',
                           subsequent_indent='  '))
                sys_exit(1)
            for sep in ':=':
                if sep in conf:
                    name, value = conf.split(sep, 1)
                    if not value:
                        argparser.error(f'Configuration {conf} without value')
                    if value == 'help':
                        value = '?'
                    helpio = StringIO()
                    eeprom.set_property(name, value, helpio)
                    helpstr = helpio.getvalue()
                    if helpstr:
                        print(fill(helpstr, initial_indent='  ',
                                   subsequent_indent='  '))
                        sys_exit(1)
                    break
            else:
                argparser.error(f'Missing name:value separator in {conf}')
        if args.vid:
            eeprom.set_property('vendor_id', args.vid)
        if args.pid:
            eeprom.set_property('product_id', args.pid)
        if args.hexdump:
            print(hexdump(eeprom.data))
        if args.hexblock is not None:
            indent = ' ' * args.hexblock
            for pos in range(0, len(eeprom.data), 16):
                hexa = ' '.join([f'{x:02x}' for x in eeprom.data[pos:pos+16]])
                print(indent, hexa, sep='')
        if args.update:
            if eeprom.commit(False, no_crc=args.full_erase):
                eeprom.reset_device()
        if args.verbose > 0:
            eeprom.dump_config()
        if args.output:
            if args.output == '-':
                eeprom.save_config(stdout)
            else:
                with open(args.output, 'wt') as ofp:
                    eeprom.save_config(ofp)

    except (ImportError, IOError, NotImplementedError, ValueError) as exc:
        print(f'\nError: {exc}', file=stderr)
        if debug:
            print(format_exc(chain=False), file=stderr)
        sys_exit(1)
    except KeyboardInterrupt:
        sys_exit(2)


if __name__ == '__main__':
    main()
